#pragma once

#include <iostream>
#include <string>
#include <memory>
#include <vector>
#include <initializer_list>

#include <functional>

//// We decide here to inline the code, i.e to write the function
//// bodies inside the class definition.  We use content_type rather
//// than std::string in the class, it is not mandatory (since
//// content_type is actually std::string as the using statement says)
//// but it will make part 3 extremely easy.

class ref_set {
public:
  using content_type = std::string;
  using ref_type     = std::shared_ptr<content_type>;

private:

  std::vector<ref_type> data; ///// This is our sets in part 1.

  // Q-211 : Make the external operator<< friend, since data is private.
  friend std::ostream& operator<<(std::ostream& os, const ref_set& rs) {
    os << '{';
    auto it = rs.data.begin();
    if(it != rs.data.end()) {
      os << *(*(it++));
      while(it != rs.data.end())
	os << ", " << *(*(it++));
    }
    os << '}';

    return os;
  }

  // Q-212 : Keep this private, it is only an internal tool.
  void remove_duplicates() {
    auto tmp = std::move(data);
    for(auto e : tmp) (*this) += e;
  }
  
public:

  ref_set()                          = default;
  ref_set(const ref_set&)            = default;
  ref_set(ref_set&&)                 = default;
  ref_set& operator=(const ref_set&) = default;
  ref_set& operator=(ref_set&&)      = default;

  // Let is keep remove_duplicate for internal use.

  // Q-213 : The data attribute is a vector, it can directly be
  // initialized witj the initializer list. Do not forget to remove
  // the eventual duplicates.
  ref_set(const std::initializer_list<ref_type>& data) : data(data) {remove_duplicates();}

  // Q-214 : These two method builds sets "out of the blue"... they
  // are thus static, no "this" in this computation.
  static ref_type make_ref(const content_type& value) {return std::make_shared<content_type>(value);}
  static ref_set empty_set() {return {};}

  // Q-215 : The first argument of the has in part 1 is now *this. Do
  // not forget to tell that it is const.
  bool has(ref_type elem) const {
    for(auto e : data) if(*e == *elem) return true;
    return false;
  }
  
  // Q-216 : The first argument of that function in part1 is now *this.
  void operator+=(ref_type elem) {if(!has(elem)) data.push_back(elem);}
  
  // Q-217 : The first argument of that function in part1 is now *this, it is const.
  ref_set operator||(const ref_set& B) const {
    ref_set res = *this;
    for(auto e : B.data) res += e;
    return res;
  }
  // Q-217 as well : same story here.
  ref_set operator&&(const ref_set& B) const {
    ref_set res;
    for(auto e : data)
      if(B.has(e)) res += e;
    return res;
  }
  
  // Q-218 : Compared with part 1, replace has(B, e) by B.has(e).
  bool operator==(const ref_set& B) const {
    if(data.size() != B.data.size()) return false;
    for(auto e : data)
      if(!B.has(e)) return false;
    return true;
  }


  // Q-22: Use the std::function type argument (passed as a const
  // reference) for passing f.
  ref_set apply(const std::function<content_type (const content_type)>& f) {
    ref_set res;
    for(auto e : data)
      res += make_ref(f(*e));
    return res;
  }
};
